(include "ssax.scm")
(use ssax-core)
(use srfi-1 extras)

(declare
  (fixnum))

;========================================================================
;		Highest-level parsers: XML to SXML
;

; procedure: SSAX:XML->SXML PORT NAMESPACE-PREFIX-ASSIG
;
; This is an instance of a SSAX parser above that returns an SXML
; representation of the XML document to be read from PORT.
; NAMESPACE-PREFIX-ASSIG is a list of (USER-PREFIX . URI-STRING)
; that assigns USER-PREFIXes to certain namespaces identified by
; particular URI-STRINGs. It may be an empty list.
; The procedure returns an SXML tree. The port points out to the
; first character after the root element.

(define (SSAX:XML->SXML port namespace-prefix-assig)
  (letrec
      ((namespaces
	(map (lambda (el)
	       (cons* #f (car el) (SSAX:uri-string->symbol (cdr el))))
	     namespace-prefix-assig))

       (RES-NAME->SXML
	(lambda (res-name)
	  (string->symbol
	   (string-append
	    (symbol->string (car res-name))
	    ":"
	    (symbol->string (cdr res-name))))))

       ; given the list of fragments (some of which are text strings)
       ; reverse the list and concatenate adjacent text strings
       (reverse-collect-str
	(lambda (fragments)
	  (if (null? fragments) '()	; a shortcut
	      (let loop ((fragments fragments) (result '()) (strs '()))
		(cond
		 ((null? fragments)
		  (if (null? strs) result
		      (cons (string-concatenate strs) result)))
		 ((string? (car fragments))
		  (loop (cdr fragments) result (cons (car fragments) strs)))
		 (else
		  (loop (cdr fragments)
			(cons
			 (car fragments)
			 (if (null? strs) result
			     (cons (string-concatenate strs) result)))
			'())))))))

       ; given the list of fragments (some of which are text strings)
       ; reverse the list and concatenate adjacent text strings
       ; We also drop "unsignificant" whitespace, that is, whitespace
       ; in front, behind and between elements. The whitespace that
       ; is included in character data is not affected.
       (reverse-collect-str-drop-ws
	(lambda (fragments)
	  (cond 
	   ((null? fragments) '())		; a shortcut
	   ((and (string? (car fragments))	; another shortcut
		 (null? (cdr fragments))	; remove trailing ws
		 (string-whitespace? (car fragments))) '())
	   (else
	    (let loop ((fragments fragments) (result '()) (strs '())
		       (all-whitespace? #t))
	      (cond
	       ((null? fragments)
		(if all-whitespace? result	; remove leading ws
		    (cons (string-concatenate strs) result)))
	       ((string? (car fragments))
		(loop (cdr fragments) result (cons (car fragments) strs)
		      (and all-whitespace?
			   (string-whitespace? (car fragments)))))
	       (else
		(loop (cdr fragments)
		      (cons
		       (car fragments)
		       (if all-whitespace? result
			   (cons (string-concatenate strs) result)))
		      '() #t))))))))
       )
    (let ((result
	   (reverse
	    ((SSAX:make-parser
	      NEW-LEVEL-SEED 
	      (lambda (elem-gi attributes namespaces
			       expected-content seed)
		'())
   
	      FINISH-ELEMENT
	      (lambda (elem-gi attributes namespaces parent-seed seed)
		(let ((seed (reverse-collect-str-drop-ws seed))
		      (attrs
		       (attlist-fold
			(lambda (attr accum)
			  (cons (list 
				 (if (symbol? (car attr)) (car attr)
				     (RES-NAME->SXML (car attr)))
				 (cdr attr)) accum))
			'() attributes)))
		  (cons
		   (cons 
		    (if (symbol? elem-gi) elem-gi
			(RES-NAME->SXML elem-gi))
		    (if (null? attrs) seed
			(cons (cons '@ attrs) seed)))
		   parent-seed)))

	      CHAR-DATA-HANDLER
	      (lambda (string1 string2 seed)
		(if (zero? (string-length string2)) (cons string1 seed)
		    (cons* string2 string1 seed)))

	      DOCTYPE
	      (lambda (port docname systemid internal-subset? seed)
		(when internal-subset?
		  (SSAX:warn port
			     "Internal DTD subset is not currently handled ")
		  (SSAX:skip-internal-dtd port))
		(SSAX:warn port "DOCTYPE DECL " docname " "
			   systemid " found and skipped")
		(values #f '() namespaces seed))

	      UNDECL-ROOT
	      (lambda (elem-gi seed)
		(values #f '() namespaces seed))

	      PI
	      ((*DEFAULT* .
			  (lambda (port pi-tag seed)
			    (cons
			     (list '*PI* pi-tag (SSAX:read-pi-body-as-string port))
			     seed))))
	      )
	     port '()))))
      (cons '*TOP*
	    (if (null? namespace-prefix-assig) result
		(cons (cons '*NAMESPACES* 
			    (map (lambda (ns) (list (car ns) (cdr ns)))
				 namespace-prefix-assig))
		      result)))
      )))

;;;; Error handler

;;; According to the SSAX convention this function
;;; accepts the port as its first argument which is used for
;;; location of the error in input file.
;;; Other parameters are considered as error messages,
;;;  they are printed to stderr as is.
(define (parser-error port msg . args)
  (let-values ([(row col) (port-position port)])
    (signal
     (make-composite-condition
      (make-property-condition 'ssax 'port port)
      (make-property-condition 'exn 'message (apply conc "[SSAX: port " (port-name port) ", at " (add1 row) "/" col "] " msg args))))))
