;========================================================================
;		Highest-level parsers: XML to SXML

(define-macro (let-values* bindings . body)
  (if (null? bindings) (cons 'begin body)
      (apply (lambda (vars initializer)
	 (let ((cont 
		(cons 'let-values* 
		      (cons (cdr bindings) body))))
	   (cond
	    ((not (pair? vars))		; regular let case, a single var
	     `(let ((,vars ,initializer)) ,cont))
	    ((null? (cdr vars))		; single var, see the prev case
	     `(let ((,(car vars) ,initializer)) ,cont))
	   (else			; the most generic case
	    `(receive ,vars ,initializer ,cont)))))
       (car bindings))))

; These parsers are a set of syntactic forms to instantiate a SSAX parser.
; A user can instantiate the parser to do the full validation, or
; no validation, or any particular validation. The user specifies
; which PI he wants to be notified about. The user tells what to do
; with the parsed character and element data. The latter handlers
; determine if the parsing follows a SAX or a DOM model.

; syntax: SSAX:make-pi-parser my-pi-handlers
; Create a parser to parse and process one Processing Element (PI).

; my-pi-handlers
;	An assoc list of pairs (PI-TAG . PI-HANDLER)
;	where PI-TAG is an NCName symbol, the PI target, and
;	PI-HANDLER is a procedure PORT PI-TAG SEED
;	where PORT points to the first symbol after the PI target.
;	The handler should read the rest of the PI up to and including
;	the combination '?>' that terminates the PI. The handler should
;	return a new seed.
;	One of the PI-TAGs may be a symbol *DEFAULT*. The corresponding
;	handler will handle PIs that no other handler will. If the
;	*DEFAULT* PI-TAG is not specified, SSAX:make-pi-parser will make
;	one, which skips the body of the PI
;	
; The output of the SSAX:make-pi-parser is a procedure
;	PORT PI-TAG SEED
; that will parse the current PI accoding to user-specified handlers.

(define-macro SSAX:make-pi-parser
  (lambda (my-pi-handlers)
  `(lambda (port target seed)
    (case target
	; Generate the body of the case statement
      ,@(let loop ((pi-handlers my-pi-handlers) (default #f))
	 (cond
	  ((null? pi-handlers)
	   (if default `((else (,default port target seed)))
	       '((else
		  (SSAX:warn port "\nSkipping PI: " target "\n")
		  (SSAX:skip-pi port)
		  seed))))
	  ((eq? '*DEFAULT* (caar pi-handlers))
	   (loop (cdr pi-handlers) (cdar pi-handlers)))
	  (else
	   (cons
	    `((,(caar pi-handlers)) (,(cdar pi-handlers) port target seed))
	    (loop (cdr pi-handlers) default)))))))))


; syntax: SSAX:make-elem-parser my-new-level-seed my-finish-element
;				my-char-data-handler my-pi-handlers

; Create a parser to parse and process one element, including its
; character content or children elements. The parser is typically
; applied to the root element of a document.

; my-new-level-seed
;	procedure ELEM-GI ATTRIBUTES NAMESPACES EXPECTED-CONTENT SEED
;		where ELEM-GI is a RES-NAME of the element
;		about to be processed.
;	This procedure is to generate the seed to be passed
;	to handlers that process the content of the element.
;	This is the function identified as 'fdown' in the denotational
;	semantics of the XML parser given in the title comments to this
;	file.
;
; my-finish-element
;	procedure ELEM-GI ATTRIBUTES NAMESPACES PARENT-SEED SEED
;	This procedure is called when parsing of ELEM-GI is finished.
;	The SEED is the result from the last content parser (or
;	from my-new-level-seed if the element has the empty content).
;	PARENT-SEED is the same seed as was passed to my-new-level-seed.
;	The procedure is to generate a seed that will be the result
;	of the element parser.
;	This is the function identified as 'fup' in the denotational
;	semantics of the XML parser given in the title comments to this
;	file.
;
; my-char-data-handler
;	A STR-HANDLER
;
; my-pi-handlers
;	See SSAX:make-pi-handler above
;

; The generated parser is a
;	procedure START-TAG-HEAD PORT ELEMS ENTITIES
;	NAMESPACES PRESERVE-WS? SEED
; The procedure must be called after the start tag token has been
; read. START-TAG-HEAD is an UNRES-NAME from the start-element tag.
; ELEMS is an instance of xml-decl::elems.
; See SSAX:complete-start-tag::preserve-ws?

; Faults detected:
;	VC: XML-Spec.html#elementvalid 
;	WFC: XML-Spec.html#GIMatch



(define-macro SSAX:make-elem-parser
  (lambda (my-new-level-seed my-finish-element
	   my-char-data-handler my-pi-handlers)
  
  `(lambda (start-tag-head port elems entities namespaces
			   preserve-ws? seed)

     (define xml-space-gi (cons SSAX:Prefix-XML
				(string->symbol "space")))

     (let handle-start-tag ((start-tag-head start-tag-head)
			    (port port) (entities entities)
			    (namespaces namespaces)
			    (preserve-ws? preserve-ws?) (parent-seed seed))
       (let-values*
	(((elem-gi attributes namespaces expected-content)
	  (SSAX:complete-start-tag start-tag-head port elems
				   entities namespaces))
	 (seed
	  (,my-new-level-seed elem-gi attributes
			      namespaces expected-content parent-seed)))
	(case expected-content
	  ((EMPTY-TAG)
	   (,my-finish-element
	    elem-gi attributes namespaces parent-seed seed))
	  ((EMPTY)		; The end tag must immediately follow
	   (SSAX:assert-token 
	    (and (eqv? #\< (SSAX:skip-S port)) (SSAX:read-markup-token port))
	    'END  start-tag-head
	    (lambda (token exp-kind exp-head)
	      (parser-error port "[elementvalid] broken for " token 
		     " while expecting "
		     exp-kind exp-head)))
	   (,my-finish-element
	    elem-gi attributes namespaces parent-seed seed))
	  (else		; reading the content...
	   (let ((preserve-ws?  ; inherit or set the preserve-ws? flag
		  (cond
		   ((assoc xml-space-gi attributes) =>
		    (lambda (name-value)
		      (equal? "preserve" (cdr name-value))))
		   (else preserve-ws?))))
	     (let loop ((port port) (entities entities)
			(expect-eof? #f) (seed seed))
	       (let-values*
		(((seed term-token)
		  (SSAX:read-char-data port expect-eof?
				       ,my-char-data-handler seed)))
		(if (eof-object? term-token)
		    seed
		    (case (xml-token-kind term-token)
		      ((END)
		       (SSAX:assert-token term-token 'END  start-tag-head
			  (lambda (token exp-kind exp-head)
			    (parser-error port "[GIMatch] broken for "
				   term-token " while expecting "
				   exp-kind exp-head)))
		       (,my-finish-element
			elem-gi attributes namespaces parent-seed seed))
		      ((PI)
		       (let ((seed 
			  ((SSAX:make-pi-parser ,my-pi-handlers)
			   port (xml-token-head term-token) seed)))
			 (loop port entities expect-eof? seed)))
		      ((ENTITY-REF)
		       (let ((seed
			      (SSAX:handle-parsed-entity
			       port (xml-token-head term-token)
			       entities
			       (lambda (port entities seed)
				 (loop port entities #t seed))
			       ,my-char-data-handler
			       seed))) ; keep on reading the content after ent
			 (loop port entities expect-eof? seed)))
		      ((START)		; Start of a child element
		       (if (eq? expected-content 'PCDATA)
			   (parser-error port "[elementvalid] broken for "
				  elem-gi
				  " with char content only; unexpected token "
				  term-token))
			   ; Do other validation of the element content
			   (let ((seed
				  (handle-start-tag
				     (xml-token-head term-token)
				     port entities namespaces
				     preserve-ws? seed)))
			     (loop port entities expect-eof? seed)))
		      (else
		       (parser-error port "XML [43] broken for "
				     term-token))))))))
	  )))
)))

; syntax: SSAX:make-parser user-handler-tag user-handler-proc ...
;
; Create an XML parser, an instance of the XML parsing framework.
; This will be a SAX, a DOM, or a specialized parser depending
; on the supplied user-handlers.

; user-handler-tag is a symbol that identifies a procedural expression
; that follows the tag. Given below are tags and signatures of the
; corresponding procedures. Not all tags have to be specified. If some
; are omitted, reasonable defaults will apply.
;

; tag: DOCTYPE
; handler-procedure: PORT DOCNAME SYSTEMID INTERNAL-SUBSET? SEED
; If internal-subset? is #t, the current position in the port
; is right after we have read #\[ that begins the internal DTD subset.
; We must finish reading of this subset before we return
; (or must call skip-internal-subset if we aren't interested in reading it).
; The port at exit must be at the first symbol after the whole
; DOCTYPE declaration.
; The handler-procedure must generate four values:
;	ELEMS ENTITIES NAMESPACES SEED
; See xml-decl::elems for ELEMS. It may be #f to switch off the validation.
; NAMESPACES will typically contain USER-PREFIXes for selected URI-SYMBs.
; The default handler-procedure skips the internal subset,
; if any, and returns (values #f '() '() seed)

; tag: UNDECL-ROOT
; handler-procedure: ELEM-GI SEED
; where ELEM-GI is an UNRES-NAME of the root element. This procedure
; is called when an XML document under parsing contains _no_ DOCTYPE
; declaration.
; The handler-procedure, as a DOCTYPE handler procedure above,
; must generate four values:
;	ELEMS ENTITIES NAMESPACES SEED
; The default handler-procedure returns (values #f '() '() seed)

; tag: DECL-ROOT
; handler-procedure: ELEM-GI SEED
; where ELEM-GI is an UNRES-NAME of the root element. This procedure
; is called when an XML document under parsing does contains the DOCTYPE
; declaration.
; The handler-procedure must generate a new SEED (and verify
; that the name of the root element matches the doctype, if the handler
; so wishes). 
; The default handler-procedure is the identity function.

; tag: NEW-LEVEL-SEED
; handler-procedure: see SSAX:make-elem-parser, my-new-level-seed

; tag: FINISH-ELEMENT
; handler-procedure: see SSAX:make-elem-parser, my-finish-element

; tag: CHAR-DATA-HANDLER
; handler-procedure: see SSAX:make-elem-parser, my-char-data-handler

; tag: PI
; handler-procedure: see SSAX:make-pi-parser
; The default value is '()
 
; The generated parser is a
;	procedure PORT SEED

; This procedure parses the document prolog and then exits to
; an element parser (created by SSAX:make-elem-parser) to handle
; the rest.
;
; [1]  document ::=  prolog element Misc*
; [22] prolog ::= XMLDecl? Misc* (doctypedec | Misc*)?
; [27] Misc ::= Comment | PI |  S
;
; [28] doctypedecl ::=  '<!DOCTYPE' S Name (S ExternalID)? S? 
;			('[' (markupdecl | PEReference | S)* ']' S?)? '>'
; [29] markupdecl ::= elementdecl | AttlistDecl
;                      | EntityDecl
;                      | NotationDecl | PI
;                      | Comment 
;


(define-macro SSAX:make-parser
  (lambda user-handlers

  ; An assoc list of user-handler-tag and default handlers
  (define all-handlers
    '((DOCTYPE .
        (lambda (port docname systemid internal-subset? seed)
	  (when internal-subset?
	      (SSAX:warn port "Internal DTD subset is not currently handled ")
	      (SSAX:skip-internal-dtd port))
	  (SSAX:warn port "DOCTYPE DECL " docname " " 
		systemid " found and skipped")
	  (values #f '() '() seed)
	  ))
      (UNDECL-ROOT .
	(lambda (elem-gi seed) (values #f '() '() seed)))
      (DECL-ROOT .
	(lambda (elem-gi seed) seed))
      (NEW-LEVEL-SEED . REQD)	; required
      (FINISH-ELEMENT . REQD)	; required
      (CHAR-DATA-HANDLER . REQD)	; required
      (PI . ())
      ))

  ; Delete an association with the tag from alist
  ; exit to cont, passing the tag, tag-association, and the list
  ; of remaining associations.
  ; It's an error if the association with the tag does not exist
  ; in alist
  (define (delete-assoc alist tag cont)
    (let loop ((alist alist) (scanned '()))
      (cond
       ((null? alist) (error "Unknown user-handler-tag: " tag))
       ((eq? tag (caar alist))
	(cont tag (cdar alist) (append scanned (cdr alist))))
       (else (loop (cdr alist) (cons (car alist) scanned))))))

  ; create an assoc list of tags and handlers
  ; based on the defaults and on the given handlers
  (define (merge-handlers declared-handlers given-handlers)
    (cond
     ((null? given-handlers)		; the arguments are exhausted...
      (cond
       ((null? declared-handlers) '())
       ((not (eq? 'REQD (cdar declared-handlers))) ; default value was given:
	(cons (car declared-handlers)		   ; use it
	      (merge-handlers (cdr declared-handlers) given-handlers)))
       (else (error "The handler for the tag " (caar declared-handlers)
		    " must be specified"))))
     ((null? (cdr given-handlers))
      (error "Odd number of arguments to SSAX:make-parser"))
     (else
      (delete-assoc declared-handlers (car given-handlers)
	  (lambda (tag value alist)
	    (cons (cons tag (cadr given-handlers))
		  (merge-handlers alist (cddr given-handlers))))))))

  (let ((user-handlers (merge-handlers all-handlers user-handlers)))

    (define (get-handler tag)
      (cond
       ((assq tag user-handlers) => cdr)
       (else (error "unknown tag: " tag))))


  `(lambda (port seed)

     ; We must've just scanned the DOCTYPE token 
     ; Handle the doctype declaration and exit to
     ; scan-for-significant-prolog-token-2, and eventually, to the
     ; element parser.
     (define (handle-decl port token-head seed)
       (or (eq? (string->symbol "DOCTYPE") token-head)
	   (parser-error port "XML [22], expected DOCTYPE declaration, found "
		  token-head))
       (assert-curr-char SSAX:S-chars "XML [28], space after DOCTYPE" port)
       (SSAX:skip-S port)
       (let-values* 
	((docname (SSAX:read-QName port))
	 (systemid
	  (and (SSAX:ncname-starting-char? (SSAX:skip-S port))
	       (SSAX:read-external-ID port)))
	 (internal-subset?
	  (begin (SSAX:skip-S port)
	    (eqv? #\[ (assert-curr-char '(#\> #\[)
					"XML [28], end-of-DOCTYPE" port))))
	 ((elems entities namespaces seed)
	  (,(get-handler 'DOCTYPE) port docname systemid
			    internal-subset? seed))
	 )
	(scan-for-significant-prolog-token-2 port elems entities namespaces
					     seed)))


     ; Scan the leading PIs until we encounter either a doctype declaration
     ; or a start token (of the root element)
     ; In the latter two cases, we exit to the appropriate continuation
     (define (scan-for-significant-prolog-token-1 port seed)
       (let ((token (SSAX:scan-Misc port)))
	 (if (eof-object? token)
	     (parser-error port "XML [22], unexpected EOF")
	     (case (xml-token-kind token)
	       ((PI)
		(let ((seed 
		       ((SSAX:make-pi-parser ,(get-handler 'PI))
			port (xml-token-head token) seed)))
		  (scan-for-significant-prolog-token-1 port seed)))
	       ((DECL) (handle-decl port (xml-token-head token) seed))
	       ((START)
		(let-values*
		 (((elems entities namespaces seed)
		   (,(get-handler 'UNDECL-ROOT) (xml-token-head token) seed)))
		 (element-parser (xml-token-head token) port elems
				 entities namespaces #f seed)))
	       (else (parser-error port "XML [22], unexpected markup "
				   token))))))


     ; Scan PIs after the doctype declaration, till we encounter
     ; the start tag of the root element. After that we exit
     ; to the element parser
     (define (scan-for-significant-prolog-token-2 port elems entities
						  namespaces seed)
       (let ((token (SSAX:scan-Misc port)))
	 (if (eof-object? token)
	     (parser-error port "XML [22], unexpected EOF")
	     (case (xml-token-kind token)
	       ((PI)
		(let ((seed 
		       ((SSAX:make-pi-parser ,(get-handler 'PI))
			port (xml-token-head token) seed)))
		  (scan-for-significant-prolog-token-2 port elems entities
						       namespaces seed)))
	       ((START)
		(element-parser (xml-token-head token) port elems
		  entities namespaces #f
		  (,(get-handler 'DECL-ROOT) (xml-token-head token) seed)))
	       (else (parser-error port "XML [22], unexpected markup "
				   token))))))


     ; A procedure start-tag-head port elems entities namespaces
     ;		 preserve-ws? seed
     (define element-parser
       (SSAX:make-elem-parser ,(get-handler 'NEW-LEVEL-SEED)
			      ,(get-handler 'FINISH-ELEMENT)
			      ,(get-handler 'CHAR-DATA-HANDLER)
			      ,(get-handler 'PI)))

     ; Get the ball rolling ...
     (scan-for-significant-prolog-token-1 port seed)
))))
